import configparser
import os
from pathlib import Path
from typing import Optional


def get_default_config() -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    config["mail"] = {
        "vmail_directory": "/var/vmail",
    }
    config["db"] = {
        "connection_string": "/var/local/mailserver.db",
        "domains_table_name": "domains",
        "users_table_name": "users",
        "aliases_table_name": "aliases",
    }
    config["users"] = {
        "insert_password_hash_prefix": True,
        "password_hash_prefix": "{BLF-CRYPT}",
    }
    config["spam"] = {
        "dkim_private_key_directory": "/var/lib/rspamd/dkim",
        "dkim_maps_path": "/etc/rspamd/dkim_selectors.map",
    }
    return config


def write_default_config(path: Path) -> configparser.ConfigParser:
    config = get_default_config()
    with path.open("w", encoding="utf-8") as fp:
        config.write(fp)
    return config


def get_config(
    config_file_path: Optional[str] = os.getenv("CONFIG_FILE"),
) -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    if config_file_path is None:
        config_file_path = "/etc/mailiness.ini"

    config_file_path = Path(config_file_path)

    if not config_file_path.exists():
        config_file_path.touch()
        config = write_default_config(config_file_path)
    else:
        config.read(str(config_file_path))
    return config


# According to https://www.daemonology.net/blog/2009-06-11-cryptographic-right-answers.html
RSA_PUBLIC_EXPONENT = 65537

# Use 2048 for maximum compatibility as at year 2022.
DKIM_KEY_SIZE = 2048



import base64
import shutil
import subprocess
from datetime import datetime
from pathlib import Path
from typing import Optional

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from mailiness import g

from . import settings

debug = getattr(g, "debug", False)


def get_default_selector():
    return datetime.now().strftime("%Y%m%d")


class DKIM:
    def __init__(self, domain: str, selector: Optional[str] = None):
        self.dkim_maps_path = g.config["spam"]["dkim_maps_path"]
        self.dkim_private_key_dir = g.config["spam"]["dkim_private_key_directory"]
        dkim_map = self.load_from_dkim_map_file()
        self.private_key = None
        if domain in dkim_map.keys():
            self.selector = dkim_map[domain]
            with (
                Path(self.dkim_private_key_dir) / Path(f"{domain}.{self.selector}.key")
            ).open("rb") as fp:
                self.private_key = serialization.load_pem_private_key(
                    fp.read(), password=None
                )
        else:
            self.generate_key()
            if selector is None:
                self.selector = get_default_selector()
            else:
                self.selector = selector

        self.domain = domain

    def generate_key(self):
        self.private_key = rsa.generate_private_key(
            public_exponent=settings.RSA_PUBLIC_EXPONENT,
            key_size=settings.DKIM_KEY_SIZE,
        )

    def private_key_as_pem(self) -> str:
        as_bytes = self.private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
        return as_bytes.decode("utf-8")

    def public_key_as_der(self) -> bytes:
        return self.private_key.public_key().public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    def load_dkim_map(self, data: str) -> dict:
        result = {}
        for line in data.splitlines():
            domain, selector = line.split(" ")
            result[domain] = selector
        return result

    def write_dkim_map(self, data: dict):
        shutil.copy2(self.dkim_maps_path, self.dkim_maps_path + ".bak")

        new_map = ""
        for domain, selector in data.items():
            new_map += f"{domain} {selector}\n"

        with Path(self.dkim_maps_path).open("w", encoding="utf-8") as fp:
            fp.write(new_map)

    def dns_txt_record(self) -> str:
        b64_der_pubkey = base64.b64encode(self.public_key_as_der())
        txt_record = ""
        txt_record += f"DNS TXT Record for {self.domain}\n\n\n\n"
        txt_record += f"Name:\n\n{self.selector}._domainkey\n\n\n"
        txt_record += f"Content:\n\nv=DKIM1;k=rsa;p={b64_der_pubkey.decode('utf-8')}"
        return txt_record

    def delete_key(self):
        dkim_map = self.load_from_dkim_map_file()
        del dkim_map[self.domain]
        self.write_dkim_map(dkim_map)

        key_file = Path(self.dkim_private_key_dir) / Path(
            f"{self.domain}.{self.selector}.key"
        )

        key_file.unlink(missing_ok=True)

        if not debug:
            self._run_system_hooks()

    def load_from_dkim_map_file(self) -> dict:
        dkim_map_file = Path(self.dkim_maps_path)

        if dkim_map_file.exists():
            with dkim_map_file.open("r", encoding="utf-8") as fp:
                return self.load_dkim_map(fp.read())
        return {}

    def save_private_key(self):
        dest = Path(self.dkim_private_key_dir) / Path(
            f"{self.domain}.{self.selector}.key"
        )

        with dest.open("w", encoding="utf-8") as fp:
            fp.write(self.private_key_as_pem())

        dkim_map = self.load_from_dkim_map_file()

        dkim_map[self.domain] = self.selector

        self.write_dkim_map(dkim_map)

        if not debug:
            self._run_system_hooks(dest)

    def _run_system_hooks(self, keyfile=None):
        subprocess.run(["systemctl", "reload", "rspamd"])
        if keyfile:
            shutil.chown(str(keyfile), user="_rspamd", group="_rspamd")



import io
import secrets
import shutil
from argparse import Namespace
from getpass import getpass
from pathlib import Path

from rich.console import Console

from mailiness import g

from . import dkim, repo

console = Console()


def handle_dkim_keygen(args: Namespace):
    key = dkim.DKIM(domain=args.domain, selector=args.selector)

    print(key.private_key_as_pem())

    if not args.quiet:
        print(key.dns_txt_record())

    if args.save:
        key.save_private_key()


def handle_dkim_show(args: Namespace):
    key = dkim.DKIM(domain=args.domain)
    print(key.private_key_as_pem())
    print(key.dns_txt_record())


def handle_domain_add(args: Namespace):
    domain_repo = repo.DomainRepository()
    tbl = domain_repo.create(args.name)
    console.print(f"{args.name} added to database")
    console.print(tbl)

    if args.dkim:
        key = dkim.DKIM(domain=args.name, selector=args.selector)
        key.save_private_key()
        print(key.dns_txt_record())


def handle_domain_edit_name(args: Namespace):
    domain_repo = repo.DomainRepository()
    tbl = domain_repo.edit("name", args.old_name, args.new_name)
    console.print(f"{args.old_name} changed to {args.new_name}")
    console.print(tbl)


def handle_domain_delete(args: Namespace):
    domain_repo = repo.DomainRepository()
    if args.yes:
        answer = "y"
    else:
        answer = input(f"Are you sure you want to delete {args.name}? (y/n)")

    while answer != "y" and answer != "n":
        answer = input(f"Are you sure you want to delete {args.name}? (y/n)")

    if answer == "y":
        domain_repo.delete(args.name)
        console.print(f"{args.name} deleted.")

        def _delete_dkim_key(domain):
            key = dkim.DKIM(domain=domain)
            key.delete_key()
            print("Private key deleted.")

        def _delete_mailbox_directory(domain):
            vmail_directory = Path(g.config["mail"]["vmail_directory"])
            shutil.rmtree(vmail_directory / domain)
            print("Mailboxes deleted.")

        if args.all:
            _delete_dkim_key(args.name)
            _delete_mailbox_directory(args.name)
        else:
            if args.dkim:
                _delete_dkim_key(args.name)
            elif args.mailbox:
                _delete_mailbox_directory(args.name)
    else:
        console.print(f"{args.name} not deleted.")


def handle_domain_list(args: Namespace):
    domain_repo = repo.DomainRepository()
    tbl = domain_repo.index()
    console.print(tbl)


def handle_user_add(args: Namespace):
    user_repo = repo.UserRepository()
    password = args.password

    if password is None:
        if args.random_password:
            password = secrets.token_urlsafe(16)
        else:
            password = getpass()
    tbl = user_repo.create(email=args.email, password=password, quota=args.quota)
    console.print(tbl)


def handle_user_edit(args: Namespace):
    user_repo = repo.UserRepository()
    password = args.password
    if password is None:
        if args.random_password:
            password = secrets.token_urlsafe(16)
        elif args.password_prompt:
            password = getpass()

    user_repo.edit(
        email=args.email, new_email=args.new_email, password=password, quota=args.quota
    )
    console.print(f"User {args.email} updated.")


def handle_user_list(args: Namespace):
    user_repo = repo.UserRepository()
    tbl = user_repo.index(domain=args.domain)
    console.print(tbl)


def handle_user_delete(args: Namespace):
    user_repo = repo.UserRepository()
    user_repo.delete(email=args.email)

    console.print(f"User {args.email} was deleted.")

    if args.mail:

        user, domain = args.email.split("@")

        vmail_directory = Path(g.config["mail"]["vmail_directory"])

        vmail_domain_directory = vmail_directory / domain

        user_vmail_directory = vmail_domain_directory / user

        shutil.rmtree(user_vmail_directory)

        console.print("User's mailbox deleted.")


def handle_alias_add(args: Namespace):
    alias_repo = repo.AliasRepository()
    tbl = alias_repo.create(args.from_address, args.to_address)
    console.print(tbl)


def handle_alias_list(args: Namespace):
    alias_repo = repo.AliasRepository()
    tbl = alias_repo.index(args.domain)
    console.print(tbl)


def handle_alias_edit(args: Namespace):
    alias_repo = repo.AliasRepository()
    tbl = alias_repo.edit(args.from_address, args.new_from, args.to)
    console.print(tbl)


def handle_alias_delete(args: Namespace):
    alias_repo = repo.AliasRepository()
    alias_repo.delete(args.from_address)


def handle_config_show(args: Namespace):
    dest = io.StringIO()
    g.config.write(dest)
    dest.seek(0)
    print(dest.read())
    dest.close()



import sqlite3
from typing import Optional, Union

import bcrypt
from rich.table import Table

from mailiness import g


def get_db_conn(dsn=g.config["db"]["connection_string"]):
    return sqlite3.connect(dsn)


class BaseRepository:
    def __init__(self, conn=None):
        if conn is None:
            self.db_conn = get_db_conn()
        else:
            self.db_conn = conn
        self.cursor = self.db_conn.cursor()
        self.data = {"headers": ("ID (rowid)", "Name"), "rows": []}

    def _prettify_data(self) -> Table:
        table = Table(title="Domains")
        for header in self.data["headers"]:
            table.add_column(header)

        for row in self.data["rows"]:
            table.add_row(*[str(col) for col in row])

        return table


class DomainRepository(BaseRepository):
    def index(self, pretty=True) -> Union[dict, Table]:
        """
        Return an domain names in the table.

        If pretty is true, return a rich table representation.
        """
        result = self.cursor.execute(
            "SELECT rowid, name FROM %s" % g.config["db"]["domains_table_name"]
        )
        self.data["rows"] = result.fetchall()
        return self._prettify_data() if pretty else self.data

    def create(self, name: str, pretty=True) -> Union[dict, Table]:
        """
        Add a domain name to the database and return the result.

        If pretty is true, return a rich table representation.
        """
        result = self.cursor.execute(
            f"INSERT INTO {g.config['db']['domains_table_name']} VALUES(?) RETURNING rowid, name",
            [name],
        )
        self.data["rows"] = result.fetchall()
        self.db_conn.commit()
        return self._prettify_data() if pretty else self.data

    def edit(self, what: str, old: str, new: str, pretty=True) -> Union[dict, Table]:
        """
        Change a domain's "what" attribute from old to new.
        """
        if what in ("name",):
            result = self.cursor.execute(
                f"UPDATE {g.config['db']['domains_table_name']} SET {what}=? WHERE {what}=? RETURNING rowid, name",
                [new, old],
            )
        else:
            raise ValueError(f"I don't know what {what} is.")
        self.data["rows"] = result.fetchall()
        self.db_conn.commit()
        return self._prettify_data() if pretty else self.data

    def delete(self, name):
        """
        Delete a domain name.
        """
        self.cursor.execute(
            f"DELETE FROM {g.config['db']['domains_table_name']} WHERE name=?", [name]
        )
        self.db_conn.commit()


class UserRepository(BaseRepository):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data["headers"] = ("ID (Row id)", "Email", "Quota (GB)")

    def _set_data(self, rows: list[tuple]):

        data = [list(row) for row in rows]

        for row in data:
            quota = int(row[2])
            row[2] = quota / 1_000_000_000
        self.data["rows"] = data

    def index(self, domain: Optional[str] = None, pretty=True) -> Union[dict, Table]:
        """
        Return a list of all users.

        If domain is provided, filter the list to show this domain's users only.
        """
        stmt = f"SELECT rowid, email, quota FROM {g.config['db']['users_table_name']}"
        if domain:
            result = self.cursor.execute(
                f"SELECT rowid FROM {g.config['db']['domains_table_name']} WHERE name=?",
                [domain],
            )
            row = result.fetchone()
            if len(row) < 0:
                raise Exception(f"Domain {domain} doesn't exist")
            domain_id = int(row[0])
            stmt += f" WHERE domain_id={domain_id}"

        result = self.cursor.execute(stmt)

        self._set_data(result.fetchall())
        return self._prettify_data() if pretty else self.data

    def _hash_password(self, password: str) -> str:
        h = bcrypt.hashpw(password.encode("utf-8"), bcrypt.gensalt())

        s = h.decode("utf-8")

        if g.config.getboolean("users", "insert_password_hash_prefix"):
            s = g.config["users"]["password_hash_prefix"] + s

        return s

    def _get_domain_id_from_email(self, email: str) -> int:
        _, domain = email.split("@")
        result = self.cursor.execute(
            f"SELECT rowid FROM {g.config['db']['domains_table_name']} WHERE name=?",
            [domain],
        )
        row = result.fetchone()
        if len(row) < 1:
            raise Exception(f"Domain {domain} doesn't exist.")

        return int(row[0])

    def _quota_gb_to_bytes(self, quota: int) -> int:
        """
        Dovecot requires quota in bytes.

        Normalize it here.
        """
        return quota * 1_000_000_000

    def create(
        self, email: str, password: str, quota: int, pretty=True
    ) -> Union[dict, Table]:
        """
        Create a new user using the parameters.

        Password will be hashed before being being stored.
        """
        hashed_password = self._hash_password(password)
        domain_id = self._get_domain_id_from_email(email)
        quota_bytes = self._quota_gb_to_bytes(quota)
        result = self.cursor.execute(
            f"INSERT INTO {g.config['db']['users_table_name']} VALUES (?,?,?,?) RETURNING rowid, email, quota",
            [domain_id, email, hashed_password, quota_bytes],
        )
        self._set_data(result.fetchall())
        self.db_conn.commit()

        return self._prettify_data() if pretty else self.data

    def edit(
        self,
        email: str,
        new_email: Optional[str] = None,
        password: Optional[str] = None,
        quota: Optional[int] = None,
    ):
        stmt = f"UPDATE {g.config['db']['users_table_name']} SET %s WHERE email=?"
        placeholders = []
        bindings = []
        if new_email:
            domain_id = self._get_domain_id_from_email(new_email)
            placeholders.append("domain_id=?")
            placeholders.append("email=?")
            bindings += [domain_id, new_email]
        if password:
            hashed_password = self._hash_password(password)
            placeholders.append("password=?")
            bindings.append(hashed_password)
        if quota:
            quota_bytes = self._quota_gb_to_bytes(quota)
            placeholders.append("quota=?")
            bindings.append(quota_bytes)

        stmt = stmt % ",".join(placeholders)

        bindings.append(email)

        self.cursor.execute(stmt, *[bindings])
        self.db_conn.commit()

    def delete(self, email: str):
        self.cursor.execute(
            f"DELETE FROM {g.config['db']['users_table_name']} WHERE email=?", [email]
        )
        self.db_conn.commit()


class AliasRepository(BaseRepository):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data["headers"] = ("ID (Row id)", "From", "To")

    def _get_domain_id(self, domain: str) -> int:
        result = self.cursor.execute(
            f"SELECT rowid FROM {g.config['db']['domains_table_name']} WHERE name=?",
            [domain],
        )
        row = result.fetchone()
        if len(row) < 1:
            raise Exception(f"Domain {domain} doesn't exist.")

        return int(row[0])

    def _get_domain_id_from_email(self, email: str) -> int:
        _, domain = email.split("@")
        return self._get_domain_id(domain)

    def index(self, domain: Optional[str] = None, pretty=True) -> Union[dict, Table]:
        """
        List all aliases.

        If domain is provided, filter the list to show only this domain's aliases.
        """
        stmt = f"SELECT rowid, from_address, to_address FROM {g.config['db']['aliases_table_name']}"
        if domain:
            domain_id = self._get_domain_id(domain)
            stmt += f" WHERE domain_id={domain_id}"

        result = self.cursor.execute(stmt)
        self.data["rows"] = result.fetchall()

        return self._prettify_data() if pretty else self.data

    def create(
        self, from_address: str, to_address: str, pretty=True
    ) -> Union[dict, Table]:
        """
        Create a new alias pointing from from_address to to_address.
        """
        domain_id = self._get_domain_id_from_email(from_address)
        result = self.cursor.execute(
            f"INSERT INTO {g.config['db']['aliases_table_name']} VALUES (?,?,?) RETURNING rowid, from_address, to_address",
            [domain_id, from_address, to_address],
        )
        self.data["rows"] = result.fetchall()
        self.db_conn.commit()

        return self._prettify_data() if pretty else self.data

    def edit(
        self,
        from_address: str,
        new_from: Optional[str] = None,
        to_address: Optional[str] = None,
        pretty=True,
    ) -> Union[dict, Table]:
        stmt = f"UPDATE {g.config['db']['aliases_table_name']} SET %s WHERE from_address=? RETURNING rowid, from_address, to_address"
        placeholders = []
        bindings = []
        if new_from:
            _, old_domain = from_address.split("@")
            _, new_domain = from_address.split("@")
            if old_domain != new_domain:
                domain_id = self._get_domain_id_from_email(new_from)
                placeholders.append("domain_id=?")
                bindings.append(domain_id)
            placeholders.append("from_address=?")
            bindings.append(new_from)

        if to_address:
            placeholders.append("to_address=?")
            bindings.append(to_address)

        stmt = stmt % ",".join([str(i) for i in placeholders])
        bindings.append(from_address)
        result = self.cursor.execute(stmt, bindings)
        self.data["rows"] = result.fetchall()
        self.db_conn.commit()

        return self._prettify_data() if pretty else self.data

    def delete(self, from_address: str):
        self.cursor.execute(
            f"DELETE FROM {g.config['db']['aliases_table_name']} WHERE from_address=?",
            [from_address],
        )
        self.db_conn.commit()



import argparse
from typing import Optional, Sequence

from mailiness import g

from . import settings

g.config = settings.get_config()

from . import commands, dkim, handlers  # noqa: E402

selector_timestamp = dkim.get_default_selector()


def add_dkim_parser(parser):
    dkim_parser = parser.add_parser("dkim", help="dkim commands")
    dkim_parser.set_defaults(func=dkim_parser.print_help, func_args=False)
    dkim_subparsers = dkim_parser.add_subparsers()
    dkim_keygen = dkim_subparsers.add_parser("keygen", help="Generate dkim key pair")
    dkim_keygen.add_argument("domain", type=str, help="example.com")
    dkim_keygen.add_argument(
        "selector",
        type=str,
        default=selector_timestamp,
        help=f"Unique string to different between different dkim keys. Example: {selector_timestamp}",
        nargs="?",
    )
    dkim_keygen.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        default=False,
        help="Don't print keys to stdout.",
    )
    dkim_keygen.add_argument(
        "--save",
        action="store_true",
        default=False,
        help="Save private key to configure directory (default: no)",
    )
    dkim_keygen.set_defaults(func=handlers.handle_dkim_keygen, func_args=True)

    dkim_show = dkim_subparsers.add_parser("show", help="display a domain's DKIM key")
    dkim_show.add_argument("domain", help="The domain whose key to show")
    dkim_show.set_defaults(func=handlers.handle_dkim_show, func_args=True)


def add_domain_parser(parser):
    domain_parser = parser.add_parser("domain", help="domain commands")
    domain_parser.set_defaults(func=domain_parser.print_help, func_args=False)
    domain_subparsers = domain_parser.add_subparsers()

    domain_add = domain_subparsers.add_parser("add", help="Add a domain name")
    domain_add.add_argument("name", help="example.com")
    domain_add.add_argument(
        "--dkim",
        action="store_true",
        default=False,
        help="Generate and save a DKIM key pair in one fell swoop.",
    )
    domain_add.add_argument(
        "--selector", nargs="?", default=selector_timestamp, help="DKIM selector"
    )
    domain_add.set_defaults(func=handlers.handle_domain_add, func_args=True)

    domain_edit = domain_subparsers.add_parser("edit", help="Edit a domain name")
    domain_edit_subparsers = domain_edit.add_subparsers()

    domain_edit_name = domain_edit_subparsers.add_parser(
        "name", help="Change the name of a domain name"
    )
    domain_edit_name.add_argument("old_name", type=str, help="Old name")
    domain_edit_name.add_argument("new_name", type=str, help="New name")
    domain_edit_name.set_defaults(func=handlers.handle_domain_edit_name, func_args=True)

    domain_delete = domain_subparsers.add_parser(
        "delete", help="Delete a domain name and all associated users and aliases"
    )
    domain_delete.add_argument("name", help="Domain name to delete")
    domain_delete.add_argument(
        "--mailbox",
        action="store_true",
        default=False,
        help="Delete mailboxes as well.",
    )
    domain_delete.add_argument(
        "--dkim", action="store_true", default=False, help="Delete DKIM key as well."
    )
    domain_delete.add_argument(
        "--all",
        action="store_true",
        default=False,
        help="Delete domain and mailboxes, DKIM keys, etc.",
    )
    domain_delete.add_argument(
        "--yes",
        "-y",
        action="store_true",
        default=False,
        help="Assume yes on all choices.",
    )
    domain_delete.set_defaults(func=handlers.handle_domain_delete, func_args=True)

    domain_list = domain_subparsers.add_parser("list", help="List domain names")
    domain_list.set_defaults(func=handlers.handle_domain_list, func_args=True)


def add_user_parser(parser):
    user_parser = parser.add_parser("user", help="user commands")
    user_parser.set_defaults(func=user_parser.print_help, func_args=False)
    user_subparsers = user_parser.add_subparsers()

    user_add = user_subparsers.add_parser("add", help="Add a user")
    user_add.add_argument("email", type=str, help="john@smith.com")
    user_add.add_argument("quota", help="The user's disk quota in GB.", type=int)
    user_add_password_group = user_add.add_mutually_exclusive_group()
    user_add_password_group.add_argument(
        "password",
        type=str,
        nargs="?",
        help="The user's password. Will be prompted if not provided",
    )
    user_add_password_group.add_argument(
        "--random-password",
        action="store_true",
        default=False,
        help="Set the user's password to a random value.",
    )
    user_add.set_defaults(func=handlers.handle_user_add, func_args=True)

    user_edit = user_subparsers.add_parser("edit", help="Edit a user")
    user_edit.add_argument("email", help="The target's current email address")
    user_edit.add_argument(
        "--new-email", "-e", help="Change the user's email address to this new one."
    )
    user_edit_password_group = user_edit.add_mutually_exclusive_group()
    user_edit_password_group.add_argument(
        "--password", "-p", help="Set the user's password to this value"
    )
    user_edit_password_group.add_argument(
        "--random-password",
        action="store_true",
        default=False,
        help="Set the user's password to a random value.",
    )
    user_edit_password_group.add_argument(
        "--password-prompt",
        action="store_true",
        default=False,
        help="Prompt for the user's new password",
    )
    user_edit.add_argument("--quota", "-q", type=int, help="Set the user's quota in GB")
    user_edit.set_defaults(func=handlers.handle_user_edit, func_args=True)

    user_list = user_subparsers.add_parser("list", help="List users")
    user_list.add_argument("--domain", "-d", help="List users for this domain only.")
    user_list.set_defaults(func=handlers.handle_user_list, func_args=True)

    user_delete = user_subparsers.add_parser("delete", help="Delete a user.")
    user_delete.add_argument("email", help="The target's email address.")
    user_delete.add_argument(
        "--mail",
        "-m",
        action="store_true",
        help="Delete the user's existing emails as well.",
    )
    user_delete.set_defaults(func=handlers.handle_user_delete, func_args=True)


def add_alias_parser(parser):
    alias_parser = parser.add_parser("alias", help="alias commands")
    alias_subparsers = alias_parser.add_subparsers()

    alias_add = alias_subparsers.add_parser("add", help="Add an alias")
    alias_add.add_argument("from_address", help="From address.")
    alias_add.add_argument("to_address", help="To address.")
    alias_add.set_defaults(func=handlers.handle_alias_add, func_args=True)

    alias_list = alias_subparsers.add_parser(
        "list", help="List all addresses pointing to this address."
    )
    alias_list.add_argument(
        "--domain", "-d", type=str, help="Show aliases from this domain only."
    )
    alias_list.set_defaults(func=handlers.handle_alias_list, func_args=True)

    alias_edit = alias_subparsers.add_parser("edit", help="Edit an alias.")
    alias_edit.add_argument("from_address", help="The target from_address.")
    alias_edit.add_argument(
        "--new-from", "-f", help="Change the from address to this one."
    )
    alias_edit.add_argument("--to", "-t", help="Change the to address to this one.")
    alias_edit.set_defaults(func=handlers.handle_alias_edit, func_args=True)

    alias_delete = alias_subparsers.add_parser("delete", help="Delete an alias")
    alias_delete.add_argument("from_address", help="The from address.")
    alias_delete.set_defaults(func=handlers.handle_alias_delete, func_args=True)


def add_config_parser(parser):
    config_parser = parser.add_parser("config", help="Configuration commands")
    config_subparsers = config_parser.add_subparsers()

    config_show = config_subparsers.add_parser(
        "show", help="Show active configuration."
    )
    config_show.set_defaults(func=handlers.handle_config_show, func_args=True)


def get_parser():
    parser = argparse.ArgumentParser(description="Manage your mail server.")
    parser.add_argument(
        "--version",
        "-v",
        action="store_true",
        default=False,
        help="Show version number.",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Work in debug mode. Most destructive actions will be prevented.",
    )
    subparsers = parser.add_subparsers()

    add_dkim_parser(subparsers)

    add_domain_parser(subparsers)

    add_user_parser(subparsers)

    add_alias_parser(subparsers)

    add_config_parser(subparsers)

    return parser


def main(args: Optional[Sequence] = None):

    parser = get_parser()

    args = parser.parse_args(args=args)

    g.debug = args.debug

    if args.version:
        commands.print_version()

    if getattr(args, "func", None):
        if args.func_args:
            args.func(args)
        else:
            args.func()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()
